@@ -5,13 +5,14 @@ import warnings
 import hashlib
 from collections import namedtuple
 from concurrent.futures import Future
-from typing import List, Optional, BinaryIO
+from typing import BinaryIO, List, Optional, Tuple
 
 import requests
 from google.cloud import storage, pubsub
 from werkzeug.datastructures import FileStorage
 
 from ..config.settings import (
+    GOOGLE_AND_OPERATOR,
     GOOGLE_DOWNLOAD_ROLE,
     GOOGLE_INTAKE_ROLE,
     GOOGLE_INTAKE_BUCKET,
@@ -24,7 +25,10 @@ from ..config.settings import (
     GOOGLE_EMAILS_TOPIC,
     GOOGLE_PATIENT_SAMPLE_TOPIC,
     GOOGLE_ARTIFACT_UPLOAD_TOPIC,
+    GOOGLE_MAX_CONDITIONAL_OPERATORS,
+    GOOGLE_MAX_DOWNLOAD_CONDITIONS,
     GOOGLE_MAX_DOWNLOAD_PERMISSIONS,
+    GOOGLE_OR_OPERATOR,
     TESTING,
     ENV,
     DEV_CFUNCTIONS_SERVER,
@@ -111,12 +115,7 @@ def grant_lister_access(user_email: str):
     """
     logger.info(f"granting list to {user_email}")
     bucket = _get_bucket(GOOGLE_DATA_BUCKET)
-
-    # Update the bucket IAM policy to include the user as a lister.
-    policy = bucket.get_iam_policy()
-    policy[GOOGLE_LISTER_ROLE] = {*policy[GOOGLE_LISTER_ROLE], f"user:{user_email}"}
-    logger.info(f"{GOOGLE_LISTER_ROLE} binding updated to {policy[GOOGLE_LISTER_ROLE]}")
-    bucket.set_iam_policy(policy)
+    grant_gcs_access(bucket, GOOGLE_LISTER_ROLE, user_email, expiring=True)
 
 
 def revoke_lister_access(user_email: str):
@@ -126,12 +125,7 @@ def revoke_lister_access(user_email: str):
     """
     logger.info(f"revoking list to {user_email}")
     bucket = _get_bucket(GOOGLE_DATA_BUCKET)
-
-    # Update the bucket IAM policy to include the user as a lister.
-    policy = bucket.get_iam_policy()
-    policy[GOOGLE_LISTER_ROLE].discard(f"user:{user_email}")
-    logger.info(f"{GOOGLE_LISTER_ROLE} binding updated to {policy[GOOGLE_LISTER_ROLE]}")
-    bucket.set_iam_policy(policy)
+    revoke_expiring_gcs_access(bucket, GOOGLE_LISTER_ROLE, user_email)
 
 
 def grant_upload_access(user_email: str):
@@ -192,7 +186,7 @@ def create_intake_bucket(user_email: str) -> storage.Bucket:
         bucket = storage_client.create_bucket(bucket)
 
     # Grant the user appropriate permissions
-    grant_expiring_gcs_access(bucket, GOOGLE_INTAKE_ROLE, user_email)
+    grant_gcs_access(bucket, GOOGLE_INTAKE_ROLE, user_email, expiring=True)
 
     return bucket
 
@@ -205,7 +199,7 @@ def refresh_intake_access(user_email: str):
     bucket = _get_bucket(bucket_name)
 
     if bucket.exists():
-        grant_expiring_gcs_access(bucket, GOOGLE_INTAKE_ROLE, user_email)
+        grant_gcs_access(bucket, GOOGLE_INTAKE_ROLE, user_email, expiring=True)
 
 
 def upload_xlsx_to_intake_bucket(
@@ -238,17 +232,16 @@ def grant_download_access(
 
     If upload_type is None, then grant access to all upload_types.
 
-    If the user already has download access for this trial and upload type, regrant the
-    permission with a new, extended TTL.
-
-    Return the GCS URI to which access has been granted.
+    If the user already has download access for this trial and upload type, idempotent.
     """
     prefix = _build_trial_upload_prefix(trial_id, upload_type)
 
     logger.info(f"Granting download access on prefix {prefix} to {user_email}")
 
     bucket = _get_bucket(GOOGLE_DATA_BUCKET)
-    grant_expiring_gcs_access(bucket, GOOGLE_DOWNLOAD_ROLE, user_email, prefix)
+    grant_gcs_access(
+        bucket, GOOGLE_DOWNLOAD_ROLE, user_email, expiring=False, prefix=prefix
+    )
 
 
 def revoke_download_access(
@@ -264,7 +257,9 @@ def revoke_download_access(
     logger.info(f"Revoking download access on {prefix} from {user_email}")
 
     bucket = _get_bucket(GOOGLE_DATA_BUCKET)
-    return revoke_expiring_gcs_access(bucket, GOOGLE_DOWNLOAD_ROLE, user_email, prefix)
+    return revoke_nonexpiring_gcs_access(
+        bucket, GOOGLE_DOWNLOAD_ROLE, user_email, prefix
+    )
 
 
 def _build_trial_upload_prefix(
@@ -276,11 +271,15 @@ def _build_trial_upload_prefix(
     return f"{trial_id}/{broad_upload_type}"
 
 
-def grant_expiring_gcs_access(
-    bucket: storage.Bucket, role: str, user_email: str, prefix: Optional[str] = None
+def grant_gcs_access(
+    bucket: storage.Bucket,
+    role: str,
+    user_email: str,
+    expiring: bool = True,
+    prefix: Optional[str] = None,
 ):
     """
-    Grant `user_email` the provided `role` on a `bucket`, expiring after `INACTIVE_USER_DAYS` 
+    Grant `user_email` the provided `role` on a `bucket`, possibly expiring after `INACTIVE_USER_DAYS` 
     days have elapsed. By default, permissions apply to the whole bucket. Optionally, provide 
     an object URL `prefix` to restrict this permission grant to only a portion of the objects 
     in the given bucket.
@@ -291,25 +290,38 @@ def grant_expiring_gcs_access(
 
     # remove the existing binding if one exists so that we can recreate it with
     # an updated TTL.
-    _find_and_pop_binding(policy, prefix, role, user_email)
-    binding = _build_binding_with_expiry(bucket.name, prefix, role, user_email)
+    _, other_conditions = _find_and_pop_binding(
+        policy, prefix, role, user_email, return_next=True
+    )
+    # will return a set of additional conditions to extend if available
+    if expiring:
+        bindings = _build_bindings_with_expiry(
+            bucket.name, [prefix], role, user_email, other_conditions=other_conditions
+        )
+    else:
+        bindings = _build_bindings_without_expiry(
+            bucket.name, [prefix], role, user_email, other_conditions=other_conditions
+        )
 
     # (re)insert the binding into the policy
-    policy.bindings.append(binding)
+    policy.bindings.extend(bindings)
     bucket.set_iam_policy(policy)
 
 
-def revoke_expiring_gcs_access(
+def revoke_nonexpiring_gcs_access(
     bucket: storage.Bucket, role: str, user_email: str, prefix: Optional[str] = None
 ):
-    """Revoke a bucket IAM policy change made by calling `grant_expiring_gcs_access`."""
+    """Revoke a bucket IAM policy change made by calling `grant_gcs_access` with expiring=False."""
     # see https://cloud.google.com/storage/docs/access-control/using-iam-permissions#code-samples_3
     policy = bucket.get_iam_policy(requested_policy_version=3)
     policy.version = 3
 
     # find and remove all matching policy bindings for this user if any exist
     for i in range(GOOGLE_MAX_DOWNLOAD_PERMISSIONS):
-        removed_binding = _find_and_pop_binding(policy, prefix, role, user_email)
+        removed_binding, other_conditions = _find_and_pop_binding(
+            policy, prefix, role, user_email, return_next=False
+        )
+
         if removed_binding is None:
             if i == 0:
                 warnings.warn(
@@ -317,6 +329,43 @@ def revoke_expiring_gcs_access(
                 )
             break
 
+        # with only return others if removed_binding is not None
+        elif len(other_conditions):
+            readd_bindings = _build_bindings_without_expiry(
+                bucket.name, [], role, user_email, other_conditions=other_conditions
+            )
+            policy.bindings.extend(readd_bindings)
+
+    bucket.set_iam_policy(policy)
+
+
+def revoke_expiring_gcs_access(
+    bucket: storage.Bucket, role: str, user_email: str, prefix: Optional[str] = None
+):
+    """Revoke a bucket IAM policy made by calling `grant_gcs_access` with (default) expiring=True."""
+    # see https://cloud.google.com/storage/docs/access-control/using-iam-permissions#code-samples_3
+    policy = bucket.get_iam_policy(requested_policy_version=3)
+    policy.version = 3
+
+    # find and remove all matching policy bindings for this user if any exist
+    for i in range(GOOGLE_MAX_DOWNLOAD_CONDITIONS):
+        removed_binding, other_conditions = _find_and_pop_binding(
+            policy, prefix, role, user_email, return_next=False
+        )
+        if removed_binding is None:
+            if i == 0:
+                warnings.warn(
+                    f"Tried to revoke a non-existent download IAM permission for {user_email}/{prefix}"
+                )
+            break
+
+        # with only return others if removed_binding is not None
+        elif len(other_conditions):
+            readd_bindings = _build_bindings_with_expiry(
+                bucket.name, [], role, user_email, other_conditions=other_conditions
+            )
+            policy.bindings.extend(readd_bindings)
+
     bucket.set_iam_policy(policy)
 
 
@@ -337,7 +386,11 @@ def revoke_all_download_access(user_email: str):
     # find and pop all download role policy bindings for this user
     for _ in range(MAX_REVOKE_ALL_ITERATIONS):
         # this finds and removes *any* download binding for the given user_email
-        if _find_and_pop_binding(policy, "", GOOGLE_DOWNLOAD_ROLE, user_email) is None:
+        # since we're popping them all, we don't care to readd any other conditions
+        if (
+            _find_and_pop_binding(policy, "", GOOGLE_DOWNLOAD_ROLE, user_email)[0]
+            is None
+        ):
             break
 
     bucket.set_iam_policy(policy)
@@ -346,12 +399,87 @@ def revoke_all_download_access(user_email: str):
 user_member = lambda email: f"user:{email}"
 
 
-def _build_binding_with_expiry(
+def _build_bindings_without_expiry(
     bucket: str,
-    prefix: Optional[str],
+    prefixes: Optional[List[str]],
+    role: str,
+    user_email: str,
+    other_conditions: List[str] = [],
+) -> List[dict]:
+    """
+    Grant the user associated with `user_email` the provided IAM `role` when acting on objects
+    in `bucket` whose URIs start with any value in `prefixes`. These permissions remains active
+    indefinitely, as the required Lister role is applied with an expiry.
+
+    GCP IAM only allows up to 12 conditional operators ie combine 13 conditions
+    Therefore must split into multiple permissions of 13 conditions each.
+    As you can only have 20 permissions total ie 19 plus Lister, errors out beyond that.
+
+    See GCP common expression language syntax overview: https://cloud.google.com/iam/docs/conditions-overview
+
+    Parameters
+    ----------
+    bucket: str
+        the name of the bucket to build the binding for
+    prefixes: Optional[List[str]]
+        a list of prefixes used to build resource.name.startsWith conditions
+        can't have more than GOOGLE_MAX_DOWNLOAD_CONDITIONS entries
+    role: str
+        the role name to build the binding for
+    user_email: str
+        the email of the user to build the binding for
+    other_conditions: List[str] = []
+        any already formatted conditions, such as returned from _find_and_pop_binding
+
+    Returns
+    -------
+    List[dict]
+        the bindings to be put onto policy.bindings
+    """
+    if len(prefixes) > GOOGLE_MAX_DOWNLOAD_CONDITIONS:
+        raise Exception(
+            f"A single user cannot have more than {GOOGLE_MAX_DOWNLOAD_CONDITIONS} download conditions"
+        )
+
+    timestamp = datetime.datetime.now()
+    return [
+        {
+            "role": role,
+            "members": {user_member(user_email)},  # convert format
+            "condition": {
+                "title": f"{role} access on {prefixes or 'bucket'}",
+                "description": f"Auto-updated by the CIDC API on {timestamp}",
+                # since this is non-expiring, all operators are OR and no brackets are needed
+                "expression": GOOGLE_OR_OPERATOR.join(
+                    [
+                        # put the other conditions in directly
+                        other_conditions.pop(0)
+                        # since we're using pop, have to stop at the end
+                        if len(other_conditions)
+                        # format object URL prefixes to the condition if specified
+                        else f'resource.name.startsWith("projects/_/buckets/{bucket}/objects/{prefixes.pop(0)}")'
+                        # can only have a certain number of operators, plus one for entries
+                        for _ in range(GOOGLE_MAX_CONDITIONAL_OPERATORS + 1)
+                        # since we're using pop, have to stop at the end
+                        if len(other_conditions) or (prefixes and len(prefixes))
+                    ]
+                ),
+            },
+        }
+        # they can only have a certain number of permissions
+        for _ in range(GOOGLE_MAX_DOWNLOAD_PERMISSIONS)
+        # since we're using pop above, have to stop at the end
+        if len(other_conditions) or (prefixes and len(prefixes))
+    ]
+
+
+def _build_bindings_with_expiry(
+    bucket: str,
+    prefixes: Optional[List[str]],
     role: str,
     user_email: str,
     ttl_days: int = INACTIVE_USER_DAYS,
+    other_conditions: List[str] = [],
 ) -> dict:
     """
     Grant the user associated with `user_email` the provided IAM `role` when acting
@@ -359,35 +487,106 @@ def _build_binding_with_expiry(
     for `ttl_days` days.
 
     See GCP common expression language syntax overview: https://cloud.google.com/iam/docs/conditions-overview
+    
+    Parameters
+    ----------
+    bucket: str
+        the name of the bucket to build the binding for
+    prefixes: Optional[List[str]]
+        a list of prefixes used to build resource.name.startsWith conditions
+        can't have more than GOOGLE_MAX_DOWNLOAD_CONDITIONS entries
+    role: str
+        the role name to build the binding for
+    user_email: str
+        the email of the user to build the binding for
+    ttl_days: int = INACTIVE_USER_DAYS
+        the number of days until this permission should expire
+    other_conditions: List[str] = []
+        any already formatted conditions, such as returned from _find_and_pop_binding
+
+    Returns
+    -------
+    List[dict]
+        the bindings to be put onto policy.bindings
     """
     timestamp = datetime.datetime.now()
     expiry_date = (timestamp + datetime.timedelta(ttl_days)).date()
-    # All conditions should have a TTL clause
-    condition = f'request.time < timestamp("{expiry_date.isoformat()}T00:00:00Z")'
-    # Add an object URL prefix to the condition if a prefix was specified
-    if prefix:
-        condition += f' && resource.name.startsWith("projects/_/buckets/{bucket}/objects/{prefix}")'
 
-    return {
-        "role": role,
-        "members": {user_member(user_email)},
-        "condition": {
-            "title": f"{role} access on {prefix or 'bucket'} until {expiry_date}",
-            "description": f"Auto-updated by the CIDC API on {timestamp}",
-            "expression": condition,
-        },
-    }
+    # going to add the expiration after, so don't return directly
+    ret = [
+        {
+            "role": role,
+            "members": {user_member(user_email)},  # convert format
+            "condition": {
+                "title": f"{role} access on {prefixes or 'bucket'}",
+                "description": f"Auto-updated by the CIDC API on {timestamp}",
+                # since this is expiring, all operators are OR
+                # if there are no entries here, we don't need brackets; so we'll deal with it later
+                "expression": GOOGLE_OR_OPERATOR.join(
+                    [
+                        # put the other conditions in directly
+                        other_conditions.pop(0)
+                        # since we're using pop, have to stop at the end
+                        if len(other_conditions)
+                        # format object URL prefixes to the condition if specified
+                        else f'resource.name.startsWith("projects/_/buckets/{bucket}/objects/{prefixes.pop(0)}")'
+                        # can only have a certain number of operators, plus one for entries but minus one for the AND
+                        # the -1 here is why this differs from the _build_bindings_without_expiry
+                        for _ in range(GOOGLE_MAX_CONDITIONAL_OPERATORS)
+                        # since we're using pop, have to stop at the end
+                        if len(other_conditions) or (prefixes and len(prefixes))
+                    ]
+                ),
+            },
+        }
+        # they can only have a certain number of permissions
+        for _ in range(GOOGLE_MAX_DOWNLOAD_PERMISSIONS)
+        # since we're using pop, have to stop at the end
+        if len(other_conditions) or (prefixes and len(prefixes))
+    ]
+
+    # Add the TTL clause
+    new_condition = f'request.time < timestamp("{expiry_date.isoformat()}T00:00:00Z")'
+    for i in range(len(ret)):  # for each binding
+        if len(ret[i]["condition"]["expression"]):
+            # if it has OR conditions, it need brackets
+            ret[i]["condition"]["expression"] = (
+                "(" + ret[i]["condition"]["expression"] + ")" + GOOGLE_AND_OPERATOR
+            )
+        # add the expiration
+        ret[i]["condition"]["expression"] += new_condition
+
+    return ret
+
+
+def _can_add_more_conditions(binding: dict) -> bool:
+    """Return if there is more space to add further prefix conditions, accounting for expiry"""
+    expression = binding.get("condition", {}).get("expression", "")
+    return (
+        expression.count(GOOGLE_OR_OPERATOR) + expression.count(GOOGLE_AND_OPERATOR)
+        < GOOGLE_MAX_CONDITIONAL_OPERATORS
+    )
 
 
 def _find_and_pop_binding(
-    policy: storage.bucket.Policy, prefix: Optional[str], role: str, user_email: str
-) -> Optional[dict]:
+    policy: storage.bucket.Policy,
+    prefix: Optional[str],
+    role: str,
+    user_email: str,
+    return_next: bool = False,
+) -> Tuple[Optional[dict], List[str]]:
     """
     Find an IAM policy binding for the given `user_email`, `policy`, and `role`, and pop
     it from the policy's bindings list if it exists.
+
+    Also returns the rest of the conditions if they exist to be readded to the bindings later.
+    If no matching binding is found and `return_next`, the last set of conditions for that
+        role/user_email if that permission can be extended.
+    In all other cases, returns an empty list.
     """
     # try to find the policy binding on the `policy`
     user_binding_index = None
+    extendable_index = None
     for i, binding in enumerate(policy.bindings):
         role_matches = binding.get("role") == role
         member_matches = binding.get("members") == {user_member(user_email)}
@@ -395,22 +594,54 @@ def _find_and_pop_binding(
             "expression", ""
         )
         if role_matches and member_matches and prefix_matches:
+            user_binding_index = i
             # a user should be a member of no more than one conditional download binding
-            if user_binding_index is not None:
+            if user_binding_index is not None or (
+                prefix is not None
+                and binding.get("condition", {}).get("expression", "").count(prefix) > 1
+            ):
                 warnings.warn(
-                    f"Found multiple conditional bindings for {user_email} on {prefix}. This is an invariant violation - "
+                    f"Found multiple conditional bindings for {user_email} on {prefix} role {role}. This is an invariant violation - "
                     "check out permissions on the CIDC GCS buckets to debug."
                 )
-                break
-            user_binding_index = i
+
+        # return the last policy binding if it doesn't have
+        elif (
+            return_next
+            and role_matches
+            and member_matches
+            and _can_add_more_conditions(binding)
+        ):
+            if extendable_index is not None:
+                warnings.warn(
+                    f"Found multiple extendable bindings for {user_email} role {role}"
+                )
+            extendable_index = i
 
     binding = (
         policy.bindings.pop(user_binding_index)
         if user_binding_index is not None
-        else None
+        else (
+            policy.bindings.pop(extendable_index)
+            if extendable_index is not None
+            else None
+        )
     )
 
-    return binding
+    # if it's an expiring permission, it'll be in the form (prefix or prefix2) and time
+    prefix_conditions = (
+        binding.get("condition", {}).get("expression", "") if binding else ""
+    )
+    if GOOGLE_AND_OPERATOR in prefix_conditions:
+        # clean up parentheses
+        prefix_conditions = prefix_conditions.split(GOOGLE_AND_OPERATOR)[0][1:-1]
+    remaining_conditions = [
+        condition
+        for condition in prefix_conditions.split(GOOGLE_OR_OPERATOR)
+        if prefix and prefix not in condition and len(condition)
+    ]
+
+    return binding, remaining_conditions
 
 
 def get_signed_url(
